//$Id: SprGEP.cc,v 1.1 2008/02/22 23:07:50 narsky Exp $


#include "StatPatternRecognition/SprExperiment.hh"
#include "StatPatternRecognition/SprGEP.hh"
#include "StatPatternRecognition/SprChromosome.hh"
#include "StatPatternRecognition/SprTrainedGEP.hh"
#include "StatPatternRecognition/SprUtils.hh"
#include "StatPatternRecognition/SprPoint.hh"
#include "StatPatternRecognition/SprAbsTwoClassCriterion.hh"

#include <stdio.h>
#include <cassert>
#include <utility>
#include <algorithm>
#include <functional>
#include <cmath>
#include <ctime>
#include <sstream>

using namespace std;

const int SprGEP::Variable = 500;
const int SprGEP::Constant = 400; 
const int SprGEP::Max_Depth = 40;


SprGEP::SprGEP(SprAbsFilter* data, 
			const SprAbsTwoClassCriterion* crit,
		    unsigned gene_head_length,
			unsigned genes_per_chromosome,
			unsigned constants_per_chromosome,
			unsigned population_size,
			int Max_Epoch,
			string functions)

  : 
  SprAbsClassifier(data),
  crit_(crit),
  gene_head_length_(gene_head_length),
  genes_per_chromosome_(genes_per_chromosome),
  constants_per_chromosome_(constants_per_chromosome),
  population_size_(population_size),
  Max_Epoch_(Max_Epoch),
  chromosomes_(population_size),
  mutation_rate_(0.05), // 0.05
  RIS_rate_(0.1), // 0.1
  IS_rate_(0.1), // 0.1
  One_Point_rate_(0.3), // 0.3
  Two_Point_rate_(0.3), // 0.3
  Whole_Gene_rate_(0.1), // 0.1
  Constant_mutation_(0.1), // 0.1 0.7
  Constant_swap_(0.1), // 0.1
  Constant_range_(10.0), // 10.0
  cls0_(0), 
  cls1_(1), 
  cut_(),
  nSorted_(0),
  sorted0_(),
  sorted1_(),
  valPrint_(0),
  division_()
{
  assert (gene_head_length > 1);
  assert (genes_per_chromosome > 0);
  assert (population_size > 1);

  cout << "Begin GEP" << endl << 
	  " Genes per Chromosome         = " << genes_per_chromosome_ << endl <<
	  " Gene head length             = " <<	gene_head_length_ << endl <<
	  " Constants per Chromosome     = " << constants_per_chromosome_ << endl <<
	  " Chromosome population size   = " << population_size_ << endl <<
      " Extra functions requested    = " << functions << endl <<
	  " Maximum Epochs/Generations   = " << Max_Epoch_ << endl;
  FunctionSet = "+-*/AQELGO";

  this->setClasses();
  this->createFunctionSet(functions);
  this->createPopulation();
}

  void SprGEP::setRates(double mutation_rate,
				double RIS_rate,
				double IS_rate,
				double One_Point_rate,
				double Two_Point_rate,
				double Whole_Gene_rate,
				double Constant_mutation,
				double Constant_swap,
				double Constant_range) {
	mutation_rate_ = mutation_rate;
	RIS_rate_ = RIS_rate;
	IS_rate_ = IS_rate;
	One_Point_rate_ = One_Point_rate;
	Two_Point_rate_ = Two_Point_rate;
	Whole_Gene_rate_ = Whole_Gene_rate;
	Constant_mutation_ = Constant_mutation;
	Constant_swap_ = Constant_swap;
	Constant_range_ = Constant_range;

	cout <<
	" Mutation Rate                = " << mutation_rate_ << endl <<
	" Root Insertion Rate          = " << RIS_rate_ << endl <<
	" Insertion Rate               = " << IS_rate_ << endl <<
	" One Point Transposition Rate = " << One_Point_rate_ << endl <<
	" Two Point Transposition Rate = " << Two_Point_rate_ << endl <<
	" Whole Gene Transpose Rate    = " << Whole_Gene_rate_ << endl <<
	" Constant Mutation Rate       = " << Constant_mutation_ << endl <<
	" Constant Swap Rate           = " << Constant_swap_ << endl <<
	" Constant Generation Range    = " << Constant_range_ << endl;
    
  }



  void SprGEP::createFunctionSet(string functions) {
	  // The functions +-*/ are always included
	FunctionsInUse.clear();
	FunctionsInUse.push_back(Plus);
	FunctionsInUse.push_back(Minus);
	FunctionsInUse.push_back(Times);
	FunctionsInUse.push_back(Divide);

	for(int i=0;i<functions.size();i++) {
		if(FunctionSet.find(functions.at(i),0) != string::npos) {
			int f = functionFromChar(functions.at(i));
			if(find(FunctionsInUse.begin(),FunctionsInUse.end(),f) == FunctionsInUse.end()) {
					FunctionsInUse.push_back(f);
			}
		} else {
				cerr << "Unknown function \"" << functions.at(i) << "\" requested." << endl;
		}
	}
	cout << "Using functions " << FunctionsInUse << endl;
	function_arg_max_ = 2;
  }

bool SprGEP::setValidation(const SprAbsFilter* valData, unsigned valPrint) {
	valData_ = valData;
	valPrint_ = valPrint;
	return true;
}

int SprGEP::functionFromChar(char f) {
	  switch(f) {
		  case '+':
			  return SprGEP::Plus;
		  case '-':
			  return SprGEP::Minus;
		  case '*':
			  return SprGEP::Times;
		  case '/':
			  return SprGEP::Divide;
		  case 'Q':
			  return SprGEP::Sqrt;
		  case 'E':
			  return SprGEP::Exp;
		  case 'L':
			  return SprGEP::LessThan;
		  case 'G':
			  return SprGEP::GreaterThan;
		  case 'A':
			  return SprGEP::Abs;
		  case 'O':
			  return SprGEP::Log;
		  default:
			  return -1;
	  }
	  return -1;
}

char SprGEP::charFromFunction(int f) {
	switch(f) {
		case SprGEP::Plus:
			return '+';
		case SprGEP::Minus:
			return '-';
		case SprGEP::Times:
			return '*';
		case SprGEP::Divide:
			return '/';
		case SprGEP::Sqrt:
			return 'Q';
		case SprGEP::Exp:
			return 'E';
		case SprGEP::LessThan:
			return 'L';
		case SprGEP::GreaterThan:
			return 'G';
		case SprGEP::Abs:
			return 'A';	
		case SprGEP::Log:
			return 'O';	
		default:
			return '!';
	}
}

vector<vector<int> > SprGene::generateTree() const {

		vector<int> geneVector(head_);
		geneVector.insert(geneVector.end(),tail_.begin(),tail_.end());

		vector<vector<int> > tree(SprGEP::Max_Depth);

		//tree.resize(SprGEP::Max_Depth);

		int nargsreq[SprGEP::Max_Depth];
		for(int i=0;i<SprGEP::Max_Depth;i++) nargsreq[i] = 0;
		int ipos = 0;
		int level = 0;
		nargsreq[level] = 1;
		while(ipos < geneVector.size()) {
			int c = geneVector.at(ipos);
			//while( level >= tree.size() ) tree.push_back(vector<int>());
			//tree.resize(level+1);
			tree[level].push_back(c);
			nargsreq[level]--;
			if(c == SprGEP::Constant || c >= SprGEP::Variable) { 
			} else {
				switch(c) {
					case SprGEP::Sqrt:
					case SprGEP::Exp:
					case SprGEP::Abs:
					case SprGEP::Log:
						nargsreq[level+1] += 1;
						break;
					case SprGEP::Plus:
					case SprGEP::Minus:
					case SprGEP::Times:
					case SprGEP::Divide:
					case SprGEP::LessThan:
					case SprGEP::GreaterThan:
						nargsreq[level+1] += 2;
						break;
					default:
						cerr << "Unrecognised function in gene '" << c << "'" << endl;
						return tree;
				}
			}
			if(nargsreq[level] == 0) {
				if(nargsreq[level+1] == 0) break;
				level++;
			}
			ipos++;
			if(level+1 >= SprGEP::Max_Depth) {
				cerr << "Max expression tree depth " << SprGEP::Max_Depth << " reached for C=";
				print();
				cout << endl;
				return tree;
			}
		}

		tree.resize(level+1);

		return tree;
}


  void SprGEP::createPopulation() {
		cout << "Creating Chromosome population of size " << population_size_ << endl;
    	chromosomes_ = std::vector<SprChromosome>(population_size_);

		unsigned size = data_->size();
		unsigned dim = data_->dim();

		unsigned head_length = gene_head_length_;
		unsigned tail_length = gene_head_length_ * (function_arg_max_ - 1) + 1;
		unsigned ch_length = gene_head_length_ + tail_length;

		cout << "Each gene will have a tail of length " << tail_length <<
			" and a total length of " << ch_length << endl;

		srand(time(0));

		// Create the chromosomes, each of 1 or more genes

		vector<SprChromosome>::iterator it = chromosomes_.begin();
		int nchromosome = 0;
		while(it != chromosomes_.end()) {
			SprChromosome &thisChromosome = *it;
			thisChromosome.setLength(genes_per_chromosome_,
				constants_per_chromosome_,
				genes_per_chromosome_*ch_length);

			for(int ch=0;ch<genes_per_chromosome_;ch++) {
				vector<int> head;
				vector<int> tail;

				for(int ih=0;ih<head_length;ih++) {
					// head can contain functions or terminals (variables or constants)
					// first position in head must be a function
					if(ih == 0) {
						int ipos = rand() % FunctionsInUse.size();
						head.push_back(FunctionsInUse.at(ipos));
					} else {
						// half the head positions will contain functions
						if(rand() % 2) {
							int ipos = rand() % FunctionsInUse.size();
							head.push_back(FunctionsInUse.at(ipos));
						} else {
							int ipos = rand() % (dim + constants_per_chromosome_);
							if(ipos >= dim) {
								head.push_back(Constant);
							} else {
								int var = Variable + ipos;
								head.push_back(var);
							}
						}
					}
				}

				for(int it=0;it<tail_length;it++) {
					int idim = rand() % (dim + constants_per_chromosome_);
					if(idim >= dim) {
						tail.push_back(Constant);
					} else {
						int var = Variable + idim;
						tail.push_back(var);
					}
				}

				thisChromosome.gene_[ch].setHead(head);
				thisChromosome.gene_[ch].setTail(tail);	

			}
			// Generate random constants for this chromosome if required
			if(constants_per_chromosome_ > 0) {
				thisChromosome.constants_.clear();
				for(int ic=0;ic<constants_per_chromosome_;ic++) {
					double val = Constant_range_ * (double) rand() / (double) RAND_MAX;
					thisChromosome.constants_.push_back(val);
				}
				//cout << "  Constants generated: " << thisChromosome.constants_ << endl;
			}
			thisChromosome.print();
			nchromosome++;
			it++;
		}

  }




void SprGEP::setClasses() 
{
  vector<SprClass> classes;
  data_->classes(classes);
  int size = classes.size();
  if( size > 0 ) cls0_ = classes[0];
  if( size > 1 ) cls1_ = classes[1];
  cout << "Classes for SprGEP are set to " 
       << cls0_ << " " << cls1_ << endl;
}

SprTrainedGEP* SprGEP::makeTrained() const {
    return new SprTrainedGEP(best_);
}


bool SprGEP::reset()
{
  cut_.clear();
  return true;
}


bool SprGEP::setData(SprAbsFilter* data)
{
  assert( data != 0 );
  data_ = data;
  sorted0_.clear();
  sorted1_.clear();
  division_.clear();
  nSorted_ = 0;
  return this->reset();
}


bool SprGEP::printValidation(unsigned cycle)
{
	// no print-out for zero training cycle
 
	if( cycle == 0 ) return true;

	double sigsig = 0;
	double sigbac = 0;
	double bacsig = 0;
	double bacbac = 0;

	// loop through validation data
  
	int vsize = valData_->size();

	for( int i=0;i<vsize;i++ ) {
		const SprPoint* p = (*valData_)[i];
		double w = valData_->w(i);
		double dclass = 0;
		if(p->class_ == cls1_) dclass = 1;

		double value = best_.Evaluate(p->x_,0);
		
		if(dclass == 1) {
			if(value > 0.) {
				sigsig += w;
			} else sigbac += w;
		} else {
			if(value <= 0.) {
				bacbac += w;
			} else bacsig += w;
		}
	}

	double fom = crit_->fom(bacbac,bacsig,sigsig,sigbac);
	
	cout << "Validation FOM=" << fom << " at Epoch " << cycle << 
		" (SS=" << sigsig << ",BB=" << bacbac << ",SB=" << sigbac << ",BS=" << bacsig <<")" << endl;
	return true;
}


bool SprGEP::train(int verbose) {


	unsigned size = data_->size();
	unsigned dim = data_->dim();

	unsigned Epoch = 0; // counts the chromosome generations

	int lastbest = 0;

	double best_fitness = 0;
	double best_loss = 1e9;
	

	while(Epoch < Max_Epoch_) {

		Epoch++;

		// For each chromosome in the population we calculate its fitness
		// by calculating its specificity and sensitivity over the whole
		// data set

		best_fitness = 0;
		best_loss = SprUtils::max();
		int best_chromosome = 0;
		int best_correct = 0;

		double sumfitness = 0;

		for(int ig=0;ig<population_size_;ig++) {

			double sumloss = 0;

			double sigsig = 0;
			double sigbac = 0;
			double bacsig = 0;
			double bacbac = 0;

			for(int ip=0; ip<size; ip++) {
				const SprPoint *p = (*data_)[ip];
				double dclass = 0;
				if(p->class_ == cls1_) dclass = 1;

			    double w = data_->w(ip);

				double value = chromosomes_[ig].Evaluate(p->x_,0);

				if(dclass == 1) {
					if(value > 0.) {
						sigsig += w;
					} else sigbac += w;
				} else {
					if(value <= 0.) {
						bacbac += w;
					} else bacsig += w;
				}

				if(dclass == 1) {
					sumloss += w*exp(-value);
				} else {
					sumloss += w*exp(value);
				}
			}

			sumloss /= (double) size;

			double fitness = crit_->fom(bacbac,bacsig,sigsig,sigbac);

			sumfitness += fitness;

			chromosomes_[ig].fitness_ = fitness;

			if(fitness > best_fitness) {
			//if(sumloss < best_loss) {
				best_fitness = fitness;
				best_loss = sumloss;
				best_chromosome = ig;
			}
		}

		double avgfitness = sumfitness / (double) population_size_;

		best_ = chromosomes_[best_chromosome];

		if(lastbest != best_chromosome) {
			lastbest = best_chromosome;	
			if(verbose > 0) {
				cout << endl << "Epoch " << Epoch << " Best chromosome " << best_chromosome << 
					" FOM " << best_fitness << " Exp.Loss " << best_loss << endl;
				best_.print();
			}
		} else if(verbose > 1) {
			cout << "Epoch " << Epoch << " Best/Average fitness " << best_fitness << "/" << avgfitness << endl;
		} else {
			//if(Epoch % 10) cout << "." ;
		}

		if(valPrint_ != 0 && Epoch % valPrint_ == 0) {
			if(!printValidation(Epoch)) {
				cerr << "Error printing validation data" << endl;
				return false;
			}
		}


		// Chromosome Mutation
		// Chromosome Transposition
		// a) by Root Insertion Sequence
		// b) by Insertion Sequence (not at the Root)
		// c) by Gene Transposition (not treated yet, and irrelevant for "+" linking function)
		// Chromosome Recombination
		// a) One Point
		// b) Two Point
		// c) Whole Gene
		// Constant mutation
		// Constant swap
		
		for(int ig=0;ig<population_size_;ig++) {
			if(ig == best_chromosome) continue; // the best chromosome is always left alone

			SprChromosome &chromosome = chromosomes_[ig];

			// Mutation
			double r = (double) (rand()) / (double) RAND_MAX;
			if(r < mutation_rate_) {
				if(verbose > 2) { cout << "Before mutation " << endl; chromosome.print();}
				bool mut = chromosome.mutate(FunctionsInUse, dim, constants_per_chromosome_);
				if(verbose > 2) { cout << "After mutation " << endl; chromosome.print();}
			}


   		    // Root insertion
			r = (double) (rand()) / (double) RAND_MAX;
			if(r < RIS_rate_) {
				if(verbose > 2) { cout << "Before RIS " << endl; chromosome.print();}
				bool ris = chromosome.RIS();
				if(verbose > 2) { cout << "After RIS " << endl; chromosome.print();}
			}
			
			// Insertion
			r = (double) (rand()) / (double) RAND_MAX;
			if(r < IS_rate_) {
				if(verbose > 2) { cout << "Before IS " << endl; chromosome.print();}
				bool is = chromosome.IS();
				if(verbose > 2) { cout << "After IS " << endl; chromosome.print();}
			}

			// One point recombination
			r = (double) (rand()) / (double) RAND_MAX;
			if(r < One_Point_rate_) {
				// Pick another chromosome from the population, randomly
				// but not the best chromosome (which must survive unchanged)
				int iother = rand() % population_size_;
				while(iother == best_chromosome || iother == ig) iother = rand() % population_size_; 
				SprChromosome &otherChromosome = chromosomes_[iother];
				if(verbose > 2) { cout << "Before One Point " << endl; chromosome.print(); otherChromosome.print();}
				bool io = chromosome.OnePoint(otherChromosome);
				if(verbose > 2) { cout << "After One Point " << endl; chromosome.print(); otherChromosome.print();}
			}

			// Two point recombination
			r = (double) (rand()) / (double) RAND_MAX;
			if(r < Two_Point_rate_) {
				// Pick another chromosome from the population, randomly
				// but not the best chromosome (which must survive unchanged)
				int iother = rand() % population_size_;
				while(iother == best_chromosome || iother == ig) iother = rand() % population_size_; 
				SprChromosome &otherChromosome = chromosomes_[iother];
				if(verbose > 2) { cout << "Before Two Point " << endl; chromosome.print(); otherChromosome.print();}
				bool it = chromosome.TwoPoint(otherChromosome);
				if(verbose > 2) { cout << "After Two Point " << endl; chromosome.print(); otherChromosome.print();}
			}

			// Whole gene recombination
			r = (double) (rand()) / (double) RAND_MAX;
			if(r < Whole_Gene_rate_) {
				// Pick another chromosome from the population, randomly
				// but not the best chromosome (which must survive unchanged)
				int iother = rand() % population_size_;
				while(iother == best_chromosome || iother == ig) iother = rand() % population_size_; 
				SprChromosome &otherChromosome = chromosomes_[iother];
				if(verbose > 2) { cout << "Before Whole Gene " << endl; chromosome.print(); otherChromosome.print();}
				bool wg = chromosome.WholeGene(otherChromosome);
				if(verbose > 2) { cout << "After Whole Gene " << endl; chromosome.print(); otherChromosome.print();}
			}

   		    // Constant mutation
			if(constants_per_chromosome_ > 0) {
				r = (double) (rand()) / (double) RAND_MAX;
				if(r < Constant_mutation_) {
					if(verbose > 2) { cout << "Before Constant Mutation " << endl; chromosome.print();}
					bool cm = chromosome.ConstantMutation(Constant_range_);
					if(verbose > 2) { cout << "After Constant Mutation " << endl; chromosome.print();}
				}
				r = (double) (rand()) / (double) RAND_MAX;
				if(r < Constant_swap_) {
					// Pick another chromosome from the population, randomly
					// but not the best chromosome (which must survive unchanged)
					int iother = rand() % population_size_;
					while(iother == best_chromosome || iother == ig) iother = rand() % population_size_; 
					SprChromosome &otherChromosome = chromosomes_[iother];
					if(verbose > 2) { cout << "Before Constant Swap " << endl; chromosome.print(); otherChromosome.print();}
					bool cs = chromosome.ConstantSwap(otherChromosome);
					if(verbose > 2) { cout << "After Constant Swap " << endl; chromosome.print(); otherChromosome.print();}
				}
			}

		}
	}

	cout << endl << "Finished training. Epoch " << Epoch << " Best chromosome FOM " << best_fitness << endl;;
	best_.print();
	if(valPrint_ != 0) {
		if(!printValidation(Epoch)) {
			cerr << "Error printing validation data" << endl;
			return false;
		}
	}


	cout << " C++ code that implements the trained chromosome: " << endl;
	best_.generateCode();


  return true;
}



char SprGEP::charFromInt(int i) {
	char c = '!';
	if(i < Constant) {
		c = charFromFunction(i);
	} else if(i == Constant) {
		c = '?';
	} else if(i >= Variable) {
		int ic = (i-Variable) % 26; // we map them all to a-z
		c = 'a' + ic;
	}
	return c;
}





void SprGEP::print(std::ostream& os) const
{
  os << "Trained GEP " << SprVersion << endl;

}
